# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003 Jrg Lehmann <joerg@luga.de>
#
# Ogg Vorbis interface by Byron Ellacott <bje@apnic.net>.
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path
import random
import string
import re
import time

import config
import MP3Info
try:
    import ogg.vorbis
    oggsupport = 1
except:
    oggsupport = 0
    pass

import events

from helper import formattime

tracknrandtitlere = re.compile("^\[?(\d+)\]? ?- ?(.*)\.(mp3|ogg)$")
UNKNOWN = "Unknown"

def capword(word):
    for i in range(len(word)):
        if word[i] not in string.punctuation: break
    else:
        i = len(word)
    return word[:i+1].upper() + word[i+1:]

def capitalize(s):
    """capitalize first character of word, leave rest intact"""
    return string.join([capword(word) for word in s.split()])

class item:
    def getname(self, songdb):
        """ return name of item """
        pass

class song(item):
    def __init__(self, path):
        assert os.path.isfile(path), "'%s' doesn't exist" % path
        self.path = os.path.normpath(path)
        self.name = os.path.basename(path)

        # guesses for title and tracknr
        match = re.match(tracknrandtitlere, self.name)
        if match:
            fntracknr = str(int(match.group(1)))
            fntitle = match.group(2)
        else:
            fntracknr = ""
            fntitle = self.name
	    if fntitle.endswith(".mp3") or fntitle.endswith(".ogg"):
	        fntitle = fntitle[:-4]

        try:
	    if self.path.endswith(".mp3"):
                self.type = "mp3"
                mp3info = MP3Info.MP3Info(open(self.path, "rb"))
                id3get = mp3info.__dict__.get
                self.title = id3get('title', "")
                self.album = id3get('album', "")
                self.artist = id3get('artist', "")
                self.year = id3get('year', "")
                self.genre  = id3get('genre', "")
                self.tracknr = id3get('track', "")
                self.length = mp3info.mpeg.length
            else:
                self.type = "ogg"
                vf = ogg.vorbis.VorbisFile(self.path)
                id3get = vf.comment().as_dict().get
                self.title = id3get('TITLE', [""])[0]
		self.title = self.title.encode('ascii', 'replace')
                self.album = id3get('ALBUM', [""])[0]
		self.album = self.album.encode('ascii', 'replace')
                self.artist = id3get('ARTIST', [""])[0]
		self.artist = self.artist.encode('ascii', 'replace')
                self.year = id3get('DATE', [""])[0]
		self.year = self.year.encode('ascii', 'replace')
                self.genre  = id3get('GENRE', [""])[0]
		self.genre = self.genre.encode('ascii', 'replace')
                self.tracknr = id3get('TRACKNUMBER', [""])[0]
		self.tracknr = self.tracknr.encode('ascii', 'replace')
                self.length = vf.time_total(0)
        except:
            self.title = fntitle
            self.album = ""
            self.artist = ""
            self.year = ""
            self.genre = ""
            self.tracknr = ""
            self.length = 0

        # do some further treatment of the song info
        
        # use title from filename, if it is a longer version of
        # the id3 tag title
        if fntitle.startswith(self.title):
	    self.title = fntitle

        # also try to use tracknr from filename, if not present as id3 tag
        if not self.tracknr:
            self.tracknr = fntracknr

        # we don't want empty album names
        if self.album=="":
            self.album = UNKNOWN

        # nor empty artist names
        if self.artist=="":
            self.artist = UNKNOWN
            
        # nor empty genres
        if self.genre=="":
            self.genre = UNKNOWN

        if not self.year or self.year=="0":
            self.year = None
        else:
            try:
                self.year = int(self.year)
            except:
                self.year = None

        # strip leading "The " in artist names, often used inconsistently
        # XXX: make this configurable
        if self.artist.startswith("The ") and len(self.artist)>4:
            self.artist = self.artist[4:]

        # XXX: make capitalization configurable
        if 1:
            self.title = capitalize(self.title)
            self.album = capitalize(self.album)
            self.artist = capitalize(self.artist)

        self.nrplayed = 0
        self.lastplayed = None
        self.rating = None

    def __cmp__(self, other):
        try:
            return cmp(self.path, other.path)
        except:
            return 1

    def __repr__(self):
        return self.path

    def getname(self, songdb):
        return self.title

    def play(self):
        self.nrplayed += 1
        self.lastplayed = time.time()
        events.hub.notify(events.songupdate(self))


class diritem(item):
    
    """ item containing other items """

    def getname(self, songdb):
        return "%s/" % self.name

    def cmpitems(self, x, y):
        """ compare to items in contained in self """
        pass
    
    def getcontents(self, songdb):
        """ return items contained in self (using database songdb) """
        pass

    def getcontentsrecursive(self, songdb):
        """ return items contained in self including subdirs (using database songdb) """
        # default implementations, which could be done more efficiently
        # by subclasses
        result = []
        for aitem in self.getcontents(songdb):
            if isinstance(aitem, diritem):
                result.extend(aitem.getcontentsrecursive(songdb))
            else:
                result.append(aitem)

        return result

    def getheader(self, item, songdb):
        """ return header (used for title bar in filelistwin) of item in self.
        The songdb is also passed
        
        Note that item can be None!
        """
        pass


class artist(diritem):
    def __init__(self, name, path=None):
        self.name = name
        self.albums = []
        self.songs = []
        self.genres = []
        self.years = []
        self.path = path
        self.rating = None

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        albums = songdb.getalbums(self.name)
        albums.sort(self.cmpitem)
        return albums

    def getheader(self, item, songdb):
        return self.name


class album(diritem):
    def __init__(self, artist, name, path=None):
        self.name = name
        self.artist = artist
        self.songs = []
        self.genres = []
        self.years = []
        self.path = path
        self.rating = None

    def cmpitem(self, x, y):
        return ( x.tracknr!="" and y.tracknr!="" and
                 cmp(int(x.tracknr), int(y.tracknr)) or
                 cmp(x.name, y.name) or 
                 cmp(x.path, y.path) )

    def getcontents(self, songdb):
        songs = songdb.getsongs(self.artist, self.name)
        songs.sort(self.cmpitem)
        return songs

    def getheader(self, item, songdb):
        return self.artist + " - " + self.name

class playlist(diritem):
    def __init__(self, path):
        assert os.path.isfile(path), "'%s' doesn't exist" % path
        self.path = os.path.normpath(path)
        self.name = os.path.basename(path)
        if self.name.endswith(".m3u"):
            self.name = self.name[:-4]
        self.songs = []

        file = open(self.path, "r")
        
        for line in file.xreadlines():
            # XXX: interpret extended m3u format (especially for streams)
            # see: http://forums.winamp.com/showthread.php?s=dbec47f3a05d10a3a77959f17926d39c&threadid=65772
            if not line.startswith("#"):
                path = line.strip()
                if not path.startswith("/"):
                    path = os.path.join(self.path, path)
                if os.path.isfile(path):
                    self.songs.append(path)
        file.close()

    def cmpitem(self, x, y):
        # this method is not very well defined for a playlist,
        # since the order of items is defined by the playlist
        # itself, but just for completeness, we set:
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        return map(songdb.getsong, self.songs)

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return self.name

#
# virtual items, i.e. items not present in database
#

class filtereditem(diritem):

    """ item, which consists of filtered items """
    
    def __init__(self, fname, item, filterfunc):
        self.fname = fname
        self.item = item
        self.filterfunc = filterfunc
        self.name = "%s <%s>" % (self.item.name, self.fname)

    def cmpitem(self, x, y):
        return self.item.cmpitem(x, y)

    def getcontents(self, songdb):
        contents = []
        for item in self.item.getcontents(songdb):
            if self.filterfunc(item):
                if isinstance(item, song):
                    contents.append(item)
                else:
                    contents.append(filtereditem(self.fname, item, self.filterfunc))
        return contents

    def getheader(self, item, songdb):
        return "%s <%s>" % (self.item.getheader(item, songdb),
                            self.fname)


class genre(diritem):
    def __init__(self, name):
        self.name = name
        self.artists = []

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def filterfunc(self, item):
        if isinstance(item, diritem):
            return self.name in item.genres
        else:
            return self.name==item.genre

    def getcontents(self, songdb):
        filteredartists = []
        for artist in songdb.getartists():
            if self.filterfunc(artist):
                filteredartists.append(filtereditem("%s=%s" % (_("Genre"),self.name),
                                                    artist,
                                                    self.filterfunc))
                
        filteredartists.sort(self.cmpitem)
        return filteredartists

    def getheader(self, item, songdb):
        return self.name

class year(diritem):
    def __init__(self, year):
        self.year = year
        self.artists = []

class decade(diritem):
    def __init__(self, decade):
        # decade = None, ..., 1960, 1970, ...
        assert decade is None or decade%10==0, \
               "decade has to be an integer multiple of 10 or None"
        if decade:
            self.decade = range(decade, decade+10)
        else:
            self.decade = decade
        self.name = decade and "%ds" % decade or _("Unknown")

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def filterfunc(self, item):
        if isinstance(item, diritem):
            if self.decade:
                for year in item.years:
                    if year in self.decade:
                        return 1
                return 0
            else:
                return None in item.years
        else:
            if item.year:
                return item.year in self.decade
            else:
                return self.decade is None

    def getcontents(self, songdb):
        filteredartists = []
        for artist in songdb.getartists():
            if self.filterfunc(artist):
                filteredartists.append(filtereditem("%s=%s" % (_("Decade"),self.name),
                                                    artist,
                                                    self.filterfunc))
                
        filteredartists.sort(self.cmpitem)
        return filteredartists

    def getheader(self, item, songdb):
        return self.name


class topplayed(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Top played songs")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        return cmp(y.nrplayed, x.nrplayed) or cmp(y.lastplayed, x.lastplayed)

    def getcontents(self, songdb):
        songs = songdb.getsongs()
        songs.sort(self.cmpitem)
        songs = [ song for song in songs[:self.maxnr]
                       if song.lastplayed is not None ]

        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return _("Top played songs")


class randomlist(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Random song list")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        # just for completeness, we set:
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        songlist = songdb.getsongs()
        songs = []

        if len(songlist)!=0:
            for i in range(self.maxnr):
                songs.append(random.choice(songlist))
        
        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return _("Random song list")


class lastplayed(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Last played songs")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        return cmp(y.lastplayed, x.lastplayed)

    def getcontents(self, songdb):
        songs = songdb.getsongs()
        songs.sort(self.cmpitem)
        songs = [ song for song in songs[:self.maxnr]
                       if song.lastplayed is not None ]
        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return _("Last played songs")


class genres(diritem):
    def getname(self, songdb):
        return "[%s (%d)]/" % (_("Genres"), len(self.getcontents(songdb)))
    
    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        genres = songdb.getgenres()
        genres.sort(self.cmpitem)
        return genres

    def getcontentsrecursive(self, songdb):
        return songdb.getsongs()

    def getheader(self, item, songdb):
        return "%s (%d)" % (_("Genres"), len(self.getcontents(songdb)))


class decades(diritem):
    def getname(self, songdb):
        return "[%s (%d)]/" % (_("Decades"), len(self.getcontents(songdb)))
    
    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        # XXX: It would be more efficient to write a songdb method
        # returning min(years), max(years).
        years = [year.year for year in songdb.getyears()]
        decades = []
        if years:
            for year in years:
                if year and year/10*10 not in decades:
                    decades.append(year/10*10)
                elif year is None and year not in decades:
                    decades.append(None)
            decades = map(decade, decades)
            
        decades.sort(self.cmpitem)
        return decades

    def getcontentsrecursive(self, songdb):
        return songdb.getsongs()

    def getheader(self, item, songdb):
        return "%s (%d)" % (_("Decades"), len(self.getcontents(songdb)))

class playlists(diritem):
    def getname(self, songdb):
        return "[%s (%d)]/" % (_("Playlists"), len(self.getcontents(songdb)))

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)
        
    def getcontents(self, songdb):
        playlists = songdb.getplaylists()
        playlists.sort(self.cmpitem)
        return playlists

    def getheader(self, item, songdb):
        return "%s (%d)" % (_("Playlists"), len(self.getcontents(songdb)))


class basedir(diritem):
    """ base dir of database view"""
    
    def __init__(self, mp3basedir, maxnr):
        self.name = _("Song Database")
        self.mp3basedir = mp3basedir
        self.maxnr = maxnr
        self.filesystembase = filesystemdir(self.mp3basedir, self.mp3basedir)
        self.topplayed = topplayed(self.maxnr)
        self.decades = decades()
        self.genres = genres()
        self.lastplayed = lastplayed(self.maxnr)
        self.randomlist = randomlist(self.maxnr)
        self.playlists = playlists()

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)
        
    def getcontents(self, songdb):
        contents = songdb.getartists()
        contents.sort(self.cmpitem)
        virtdirs = [self.filesystembase,
                    self.decades,
                    self.genres,
                    self.topplayed,
                    self.lastplayed,
                    self.randomlist,
                    self.playlists]
        if config.virtualdirectoriesattop:
            contents[:0] = virtdirs
        else:
            contents[len(contents):] = virtdirs
        return contents

    def getheader(self, item, songdb):
        return "%s (%d %s)" % ( _("Song Database"),
                                len(songdb),
                                _("entries"))


class filesystemdir(diritem):

    """ diritem corresponding to directory in filesystem """

    def __init__(self, mp3basedir, dir):
        self.mp3basedir = mp3basedir
        self.dir = dir
        
        if self.dir==self.mp3basedir:
            self.name = "[%s]" % _("Filesystem")
        else:
            self.name = self.dir[len(self.mp3basedir):].split("/")[-1]

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        items = []
        try:
            for name in os.listdir(self.dir):
                try:
                    path = os.path.join(self.dir, name)
                    if os.path.isdir(path) and os.access(path, os.R_OK|os.X_OK):
                        newitem = filesystemdir(self.mp3basedir, path)
                        items.append(newitem)
                    elif name.endswith(".mp3") and os.access(path, os.R_OK):
                        newsong = songdb.queryregistersong(path)
                        items.append(newsong)
		    elif name.endswith(".ogg") and os.access(path, os.R_OK):
                        newsong = songdb.queryregistersong(path)
                        items.append(newsong)
                except (IOError, OSError) : pass
        except OSError:
            return None
        items.sort(self.cmpitem)
        return items

    def getcontentsrecursive(self, songdb):
        # we better deny doing this
        return []

    def getheader(self, item, songdb):
        if self.dir==self.mp3basedir:
            return _("Filesystem")
        else:
            return self.name


    
